with Ada.Calendar,
     Ada.Text_Io;
package body Log is

  F : Ada.Text_Io.File_Type;

  procedure Initialize(X : in out Log_Objects) is
  begin
    begin
      Ada.Text_Io.Create(F, Ada.Text_Io.Out_File, "log");
    exception
      when Ada.Text_Io.Use_Error =>  -- if already exists, append
        Ada.Text_Io.Open(F, Ada.Text_Io.Append_File, "log");
    end;
    Ada.Text_Io.Put_Line(F,
      Duration'image(Ada.Calendar.Seconds(Ada.Calendar.Clock))
      & "--- opening log");
    Ada.Text_Io.Flush(F);
  end Initialize;

  procedure Finalize(X : in out Log_Objects) is
  begin
    if Ada.Text_Io.Is_Open(F) then
      Ada.Text_Io.Put_Line(F,
        Duration'image(Ada.Calendar.Seconds(Ada.Calendar.Clock))
        & "--- closing log");
      Ada.Text_Io.Close(F);
    end if;
  end Finalize;

  protected Log_Guard is
    entry Grab;
    procedure Release;
  private
    In_Use : Boolean := False;
  end Log_Guard;

  protected body Log_Guard is
    entry Grab when not In_Use is
    begin In_Use := True;end Grab;
    procedure Release is
    begin
      In_Use := False;
    end Release;
  end Log_Guard;

  The_Log : Log_Objects;

  procedure Note(S : in String) is
  begin
    Log_Guard.Grab;
    Ada.Text_Io.Put_Line(F,
      Duration'image(Ada.Calendar.Seconds(Ada.Calendar.Clock)) & " " & S);
    Ada.Text_Io.Flush(F);
    Log_Guard.Release;
  exception
    when others =>
      Log_Guard.Release;
  end Note;

end Log;
