
with Ada.Characters.Handling;
package body Str_Sto is

  -- Data structure initialized during package elaboration.
  type Data_Indices is range 1 .. 10_000_000;
  type Data_Strings is array(Data_Indices range <>) of Character;
  Data : Data_Strings(Data_Indices);

  Last_List : array(String_Counts) of Data_Indices;
  Next_String : String_Indices := String_Indices'first;

  function Count return String_Counts is
  -- return number of stored strings.
  begin return Next_String-1;end Count;

  function Add(S : in String) return String_Indices is
  -- Add to store. Return new index.
  -- Raise Overflow if too many or too long strings.
    Start : Data_Indices;
  begin
    if Next_String = String_Indices'last
    or else Last_List(Next_String-1) > Data_Indices'last - S'length then
      raise Overflow;
    end if;
    Start := Last_List(Next_String-1)+1;
    Last_List(Next_String) := Start+S'length-1;
    Data(Start .. Last_List(Next_String)) := Data_Strings(S);
    Next_String := Next_String+1;
    return Next_String-1;
  end Add;

  function Get(Index : in String_Indices) return String is
  -- raise Nonesuch if no such string.
    pragma suppress(range_check);
    pragma suppress(index_check);
    Start : Data_Indices;
  begin
    if Index >= Next_String then raise Nonesuch;end if;
    Start := Last_List(Index-1)+1;
    declare
      Result : String(1 .. Integer(Last_List(Index)-Start+1));
      -- get indices right
--    Piece  : constant Data_Strings(1 .. Last_List(Index)-Start+1)
--      := Data(Start .. Last_List(Index));
    begin
--    return String(Piece);
      for i in Result'range loop
        Result(i) := Data(Data_Indices(i)+Start-1);
      end loop;
      return Result;
    end;
  end Get;

  function Get_Length(Index : in String_Indices) return Natural is
  -- raise Nonesuch if no such string.
    Start : Data_Indices;
  begin
    if Index >= Next_String then raise Nonesuch;end if;
    Start := Last_List(Index-1)+1;
    return Integer(Last_List(Index)-Start+1);
  end Get_Length;

  function To_Upper(S : in Data_Strings) return String is
    Result : String(1 .. S'length);
  begin
    for i in S'range loop
      Result(Integer(i-S'first)+1) := Ada.Characters.Handling.To_Upper(S(i));
    end loop;
    return Result;
  end To_Upper;

  Upcase : array(character) of character;

  function Compare(Index : in String_Indices; S : in String)
  return Comparisons is
    use Ada.Characters.Handling;
    SI : Data_Indices;
    Len : Natural;
    j : Integer := S'first;
    UD : Character;
    US:  Character;
  begin
    if Index >= Next_String then raise Nonesuch;end if;
    SI := Last_List(Index-1)+1;
    Len := Natural(Last_List(Index) - Last_List(Index-1));
    if Len = S'length
    and then Data_Strings(S) = Data(SI .. Last_List(Index)) then
      return EQ;
    end if;
    if S'length = 0 then return GT;end if;
    if Len = 0 then return LT;end if;
    -- not EQ and neither is zero length
    if Upcase(Data(SI)) > Upcase(S(S'first)) then return GT;
    elsif Upcase(Data(SI)) < Upcase(S(S'first)) then return LT;
    else
      for di in SI .. Last_List(Index) loop
        if j > S'last then return GT;end if;
        UD := Upcase(Data(di));
        US := Upcase(S(j));
        if UD > US then return GT;
        elsif UD < US then return LT;
        end if;
        j := j+1;
      end loop;
      if j <= S'last then return LT;end if;
      return EQ;
    end if;
  end Compare;

  function Compare(Left, Right : in String_Indices)
  return Comparisons is
    use Ada.Characters.Handling;
    LI,RI : Data_Indices;
    LLen,RLen : Natural;
  begin
    if Left >= Next_String or Right >= Next_String then
      raise Nonesuch;
    end if;
    if Left = Right then return EQ;end if;
    LI := Last_List(Left-1)+1;
    LLen := Natural(Last_List(Left) - Last_List(Left-1));
    RI := Last_List(Right-1)+1;
    RLen := Natural(Last_List(Right) - Last_List(Right-1));
    if LLen = RLen
    and then To_Upper(Data(LI .. Last_List(Left)))
             = To_Upper(Data(RI .. Last_List(Right))) then
      return EQ;
    end if;
    if RLen = 0 then return GT;end if;
    if LLen = 0 then return LT;end if;
    -- not EQ and neither is zero length
    if To_Upper(Data(LI)) > To_Upper(Data(RI)) then return GT;
    elsif To_Upper(Data(LI)) < To_Upper(Data(RI)) then return LT;
    elsif To_Upper(Data(LI .. Last_List(Left)))
          > To_Upper(Data(RI .. Last_List(Right))) then
      return GT;
    elsif To_Upper(Data(LI .. Last_List(Left)))
          < To_Upper(Data(RI .. Last_List(Right))) then
      return LT;
    else return EQ;
    end if;
  end Compare;

begin
  Last_List(String_Indices'first-1) := Data'first;
  Data(Data'first) := ' ';
  Next_String := String_Indices'first;
  for c in Upcase'range loop Upcase(c) := c;end loop;
  for c in character range 'a' .. 'z' loop
    Upcase(c) := Ada.Characters.Handling.To_Upper(c);
  end loop;
end Str_Sto;
