with Ada.Characters.Handling;

package body Te_Text_Line_Sockets is
  use Ada.Characters.Handling;

  procedure Set_Encoding(Socket   : in out Socket_Type;
                         Encoding : in     Encoding_Type) is
  begin
    Socket.Encoding := Encoding;
    Socket.Remaining_In_Chunk := 0;
    Socket.Last_Chunk_Seen := False;
  end Set_Encoding;

  Hex_Val : constant array(Character) of Natural
    := ('0' => 0, '1' => 1, '2' => 2, '3' => 3, '4' => 4,
        '5' => 5, '6' => 6, '7' => 7, '8' => 8, '9' => 9,
        'a' =>10, 'b' =>11, 'c' =>12, 'd' =>13, 'e' =>14, 'f' =>15,
        'A' =>10, 'B' =>11, 'C' =>12, 'D' =>13, 'E' =>14, 'F' =>15,
        others => 0);

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0) is
    use type Claw.Natural_Int;
    Chunk_Size : Claw.Natural_Int;
    Size_Chunk_Buffer : String(1 .. 40); -- allow up to FFFFFFFFcrlf
    SZ_Last : Natural;                   -- +30 chars of chunk extension
  begin
    Last := Line'first-1;
    if Line'length = 0 then
      return;
    end if;
    case Socket.Encoding is
      when Plain =>
        Text_Line_Sockets.Get_Line(Text_Line_Sockets.Socket_Type(Socket),
                                   Line, Last, Timeout);
      when Chunked =>
        if Socket.Last_Chunk_Seen then
          return;
        end if;
        -- form is hexnumber crlf data crlf repeat, where data'length=number
        if Socket.Remaining_In_Chunk = 0 then
          -- time to read hexnumber cr lf
          Text_Line_Sockets.Get_Line(Text_Line_Sockets.Socket_Type(Socket),
                                     Size_Chunk_Buffer, SZ_Last, Timeout);
          Chunk_Size := 0;
          for si in 1 .. SZ_Last loop
            exit when not Is_Hexadecimal_Digit(Size_Chunk_Buffer(si));
            Chunk_Size := 16 * Chunk_Size + Claw.Natural_Int(Hex_Val(Size_Chunk_Buffer(si)));
          end loop;
          if Chunk_Size = 0 then
            Socket.Last_Chunk_Seen := True;
            return;
          end if;  --
          Socket.Remaining_In_Chunk := Chunk_Size+2;
        -- True chunk size is specified size + cr and lf
        end if;
        -- read data through a crlf
        Text_Line_Sockets.Get_Line(Text_Line_Sockets.Socket_Type(Socket),
                                   Line, Last, Timeout);
        -- if all is well, we've read through the crlf on a line of data.
        -- There may be more lines in the same chunk, or the crlf may be
        -- the one marking end-of-chunk, in which case we'll need to toss
        -- it, read the chunk control stuff, and continue reading the
        -- rest of the logical line.
        if Claw.Natural_Int(Last - Line'first+1) > Socket.Remaining_In_Chunk then
          Socket.Remaining_In_Chunk := 0;  -- no crlf ending this chunk!
          raise Bad_Chunk;
        end if;
        Socket.Remaining_In_Chunk
          := Socket.Remaining_In_Chunk - Claw.Natural_Int(Last - Line'first+1);
        if Socket.Remaining_In_Chunk = 0 then   -- we read up through crlf
          if Last - Line'first+1 >= 2
          and then Line(Last-1) = ascii.cr
          and then Line(Last) = ascii.lf then
            Last := Last-2;                   -- drop crlf and continue reading
            Get_Line(Socket, Line(Last+1 .. Line'last), Last, Timeout);
          else
            raise Bad_Chunk;   -- chunk didn't end with crlf!
          end if;
        end if;
    end case;
  end Get_Line;

  procedure Close(Socket : in out Socket_Type) is
  begin
    Socket.Encoding := Plain;
    Text_Line_Sockets.Close(Text_Line_Sockets.Socket_Type(Socket));
  end Close;

end Te_Text_Line_Sockets;
