with Ada.Strings.Fixed;

package body Text_Line_Sockets is

  Eol_Char: constant String(1 .. 1) := (1 => Ascii.Lf);

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0) is
    use Ada.Strings.Fixed;
    Si      : Natural;
    N_Avail,
    N_To_Get: Natural;

    procedure Get_More is
    begin
      Socket.Next := 1;
      Socket.Last := 0;
      Get(Socket, Timeout, Socket.Buffer, Socket.Last);
    exception
      when We   : Claw.Windows_Error =>
        Socket.Last := 0;
        Socket.Status := Problem; -- next Get will fail
        Ada.Exceptions.Save_Occurrence(Source => We,
                                       Target => Socket.Occurrence);
      when Oops : Claw.Not_Found_Error =>
        -- socket has been closed, treat as timeout
        Socket.Status := Eof;
        Ada.Exceptions.Save_Occurrence(Source => Oops,
                                       Target => Socket.Occurrence);
    end Get_More;
  begin
    Last := Line'First - 1;
    case Socket.Status is
      when Ok =>
        null;
      when Eof =>
        Socket.Status := Ok; -- so next read will work (if socket re-opened)
        raise Claw.Not_Found_Error;
      when Problem =>
        Socket.Status := Ok; -- to allow conversation to continue
        Ada.Exceptions.Reraise_Occurrence(Socket.Occurrence);
    end case;
    loop
      if Last >= Line'Last then
        return;
      end if;
      if Last >= Line'First and then Line(Last) = Ascii.Lf then
        return;
      end if;
      Si := Index(Socket.Buffer(Socket.Next .. Socket.Last), Eol_Char);
      if Si = 0 then
        N_Avail := Socket.Last - Socket.Next + 1;
      else
        N_Avail := Si - Socket.Next + 1;
      end if;
      N_To_Get := Integer'Min(N_Avail, Line'Last - Last);
      if N_To_Get > 0 then
        Line(Last + 1 .. Last + N_To_Get)
          := Socket.Buffer(Socket.Next .. Socket.Next + N_To_Get - 1);
        Socket.Next := Socket.Next + N_To_Get;
        Last := Last + N_To_Get;
      else
        -- n_avail=0 in socket buffer
        if Socket.Status = Ok then
          Get_More;
          if Socket.Last = 0 then
            return; -- timeout or close, return with whatever we've got
          end if;
        else
          -- can't read any more, return with whatever we've got
          return;
        end if;
      end if;
    end loop;
  end Get_Line;

  procedure Close(Socket  : in out Socket_Type) is
  begin
    Socket.Last := 0;
    Socket.Next := 1;
    Socket.Status := Ok;
    Claw.Sockets.Close(Claw.Sockets.Socket_Type(Socket));
  end Close;
end Text_Line_Sockets;
