with Claw,
     Claw.Canvas.Ram,
     Claw.Capabilities,
     Claw.Draw,
     Claw.Window_Operations;

package body Tty_Out is

  Tty_Count: Natural := 0;

  procedure New_Page(Tty : in out Tty_Type) is
  begin
    if Tty.State = Closed then
      return;
    end if;
    Tty.Length := 0;
    if Tty.State = Not_Yet_Opened then
      Tty_Count := Tty_Count + 1;
      Create(Tty, "TTY" & Integer'Image(Tty_Count));
      Show(Tty, Claw.Codes.Show_Startup);
      Tty.State := Displaying;
    end if;
    Claw.Window_Operations.Invalidate(Tty);
  exception
    when Claw.Not_Valid_Error =>
      Tty.State := Closed;
  end New_Page;

  procedure Put(Tty  : in out Tty_Type;
                Text : in     String) is
  begin
    if Tty.State = Closed then
      return;
    end if;
    if Text'Length = 0 then
      return;
    end if;
    if Tty.Length + Text'Length > Max_Length then
      Tty.Text := Tty.Text(Text'Length + 1 .. Max_Length) & Text;
      Tty.Length := Max_Length;
    else
      Tty.Text(Tty.Length + 1 .. Tty.Length + Text'Length) := Text;
      Tty.Length := Tty.Length + Text'Length;
    end if;
    if Tty.State = Not_Yet_Opened then
      Tty_Count := Tty_Count + 1;
      Create(Tty, "TTY" & Integer'Image(Tty_Count));
      Show(Tty, Claw.Codes.Show_Startup);
      Tty.State := Displaying;
    end if;
    Claw.Window_Operations.Invalidate(Tty);
  exception
    when Claw.Not_Valid_Error =>
      Tty.State := Closed;
  end Put;

  procedure Put_Line(Tty  : in out Tty_Type;
                     Text : in     String) is
  begin
    Put(Tty, Text & Ascii.Cr & Ascii.Lf);
  end Put_Line;

  procedure Close(Tty : in out Tty_Type) is
  begin
    Destroy(Tty);
    Tty.State := Closed;
  exception
    when Claw.Not_Valid_Error =>
      Tty.State := Closed; -- in case it's already closed
  end Close;

  function Is_Closed(Tty : Tty_Type) return Boolean is
  begin
    return Tty.State = Closed;
  end Is_Closed;

  procedure When_Draw(Tty     : in out Tty_Type;
                      Easel   : in out Claw.Canvas.Basic_Canvas_Type'Class;
                      Erase_Background: in Boolean;
                      Area    : in     Claw.Rectangle_Type) is
    use type Claw.Int;
    Text_Area: constant Claw.Rectangle_Type := Client_Area(Tty);
    Text_Metrics: Claw.Capabilities.Text_Metrics_Type;
    Height  : Claw.Int;
    Tail_Line_Count,
    Max_Line_Count: Claw.Int;
    Start_Showing: Positive;
    Y       : Claw.Int;
  begin
    Claw.Capabilities.Get_Text_Metrics(Easel, Text_Metrics);
    Height := Text_Metrics.Height + Text_Metrics.External_Leading;
    Max_Line_Count := (Text_Area.Bottom - Text_Area.Top) / Height;
    Tail_Line_Count := 0;
    Start_Showing := 1;
    for I in reverse 1 .. Tty.Length - 2 loop
      -- ignore trailing crlf
      if Tty.Text(I) = Ascii.Cr and Tty.Text(I + 1) = Ascii.Lf then
        Tail_Line_Count := Tail_Line_Count + 1;
        if Tail_Line_Count = Max_Line_Count then
          Start_Showing := I + 2;
          exit;
        end if;
      end if;
    end loop;
    Y := 0;
    Show_Tail:
    while Start_Showing <= Tty.Length loop
      for I in Start_Showing .. Tty.Length loop
        if I = Tty.Length then
          Claw.Draw.Put(Easel, (0, Y), Tty.Text(Start_Showing .. Tty.Length));
          exit Show_Tail;
        elsif Tty.Text(I) = Ascii.Cr and Tty.Text(I + 1) = Ascii.Lf then
          Claw.Draw.Put(Easel, (0, Y), Tty.Text(Start_Showing .. I - 1));
          Y := Y + Height;
          Start_Showing := I + 2;
          exit;
        end if;
      end loop;
    end loop Show_Tail;
  end When_Draw;

  procedure When_Close(Tty : in out Tty_Type) is
  begin
    Tty.State := Closed;
    Claw.Basic_Window.When_Close(Claw.Basic_Window.Basic_Window_Type(Tty));
  end When_Close;

end Tty_Out;
