with Claw.Basic_Window,
     Claw.Canvas;
package TTY_Out is

  -- Displays text in TTY fashion in a window.  Scrolls as necessary
  -- if the window is not big enough to hold all the text.  The
  -- first Put or Put_Line on the TTY_Type object will open it to
  -- default size and position, which the user may then change.
  -- If the user closes the window, future calls to Put or Put_Line
  -- will be nop's.

  Max_Length : constant := 5_000;  -- Max total text length

  type TTY_Type is new Claw.Basic_Window.Basic_Window_Type with private;

  procedure Put(TTY  : in out TTY_Type;
                Text : in     String);

  procedure Put_Line(TTY  : in out TTY_Type;
                     Text : in     String);

  procedure New_Page(TTY  : in out TTY_Type);

  procedure Close(TTY : in out TTY_Type);

  function Is_Closed(TTY : in TTY_Type) return Boolean;

private

  type States is (Not_Yet_Opened, Displaying, Closed);

  type TTY_Type is new Claw.Basic_Window.Basic_Window_Type with record
    State  : States := Not_Yet_Opened;
    Length : Natural :=0;
    Text   : String(1 .. Max_Length);
  end record;

  procedure When_Draw (TTY    : in out TTY_Type;
                       Easel  : in out Claw.Canvas.Basic_Canvas_Type'Class;
                       Erase_Background : in Boolean;
                       Area   : in Claw.Rectangle_Type);

  procedure When_Close (TTY : in out TTY_Type);

end TTY_Out;
